/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the UDP data structures */
#include "pxe_udp.h"

/* Define the PXE UNDI data structures */
#include "pxe_undi.h"

/* Define the PXENV+ data structures */
#include "pxenv.h"

/* Define the internal Nilo data structures and constants */
#include "nilo.h"

#include <oskit/net/ether.h>
#include <oskit/types.h>

static u_int8_t bcastaddr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

extern char packet[ETH_MAX_PACKET];
extern oskit_size_t packetlen;

/* Define Nilo's UDP data structures and constants */
#include "udp.h"

/* Define Nilo's ARP data structures  and constants */
#include "arp.h"

/* pxenv_udp_open() - open the UDP connection. There can only be
 *      one open UDP connection at a time.
 *
 * parameters:
 *     none
 *
 * returns:
 *     udpopen.status - the exit status of this function. This is set
 *                       to one of the PXENV_STATUS_xxx constants.
 *     udpopen.src_ip - the IP number of this node.
 */
pxenv_ret_code_t
pxenv_udp_open (pxenv_udp_open_t *udpopen)
{
    NILO_TRACE_FUNCTION;

    udpopen->status = PXENV_STATUS_UDP_OPEN;

    /* Send a DHCPDISCOVER packet. This gets our IP number. */
    dhcp_handler(client_info->xid = getxid());

    if (client_info->client_state == DISCOVERING)
        panic ("DHCP session is not complete.\r\n");

    /* We don't have to do much but set the IP number of ourself */
    udpopen->src_ip.num = client_info->yiaddr.num;

    /* Send a DHCPREQUEST packet */
    dhcp_handler((long)client_info->xid = getxid());

    NILO_TRACE_MSG ("DHCP session done.\r\n");

    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_udp_close() - close the currently open UDP connection.
 *
 * parameters:
 *     none
 * returns:
 *     udpwrite.status - the exit status of this function. This is set
 *                       to one of the PXENV_STATUS_xxx constants.
 */
pxenv_ret_code_t
pxenv_udp_close (pxenv_udp_close_t *udpclose)
{
    NILO_TRACE_FUNCTION;

    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_udp_read() - read one packet using an already opened UDP
 *      connection.
 *
 * parameters:
 *     udpread.dest_ip - if non-zero, filter all incoming
 *                packets and receive the ones for this IP address. If
 *                UDP_READ_ANY_IP (zero), then read any packet.
 *     udpread.d_port - if non-zero, filter all incoming packets
 *               and receive the ones for this UDP port. If
 *               UDP_READ_ANY_PORT (zero), then read any packet.
 *     udpread.buffer_size - size of the receive buffer.
 *     udpread.buffer - the segment the buffer is in.
 *
 * returns:
 *     udpwrite.status - the exit status of this function. This is set
 *                       to one of the PXENV_STATUS_xxx constants.
 *     udpread.src_ip - The IP number of the sending node.
 *     udpread.dest_ip - If this parameter was UDP_READ_ANY_IP
 *                      (zero), then return the destination IP number.
 *     udpread.s_port - The UDP port number of the sending node. 
 *     udpread.d_port - If this parameter was zero, then
 *                        return the  destination UDP port.
 *     udpread.buffer_size - Number of bytes written into
 *                        the receive buffer.
 */
pxenv_ret_code_t
pxenv_udp_read (pxenv_udp_read_t *udpread)
{
    struct iphdr *ip;
    struct udphdr *udp;
    arp_t *arp_reply;
    unsigned long time;
    u_int32_t num;
    
    NILO_TRACE_FUNCTION;
    
    //    NILO_TRACE_MSG ("packet size is 0x%02x\r\n", packetlen);
    
    udpread->buffer_size = packetlen;
    time = currticks() + 50;
    while(time > currticks()) {
      if ((num = eth_poll())) {	/* We have something! */
	
	/* See if it's an ARP reply */
#if 0
	if ((packetlen >= ETHER_HDR_SIZE + sizeof(arp_t)) && 
#else
	    if (packetlen == 60 &&
		(((packet[12] << 8) | packet[13]) ==  ETHERTYPE_ARP)) {
#endif
	  arp_reply = (arp_t *)&packet[ETHER_HDR_SIZE];
	  NILO_TRACE_MSG ("Got an ARP packet.\r\n");
	if (ntohs(arp_reply->hwtype) == 1) {
	  udpread->buffer = packet[ETHER_HDR_SIZE];
//	  udpread->buffer_size = sizeof(arp_t);
	  memcpy(&num, arp_reply->sipaddr, sizeof(num));
	  if (lookup_arp (client_info->siaddr) == 0) {
	    add_to_cache((ip4_t)num, arp_reply->shwaddr,
			 ARP_SERVER);
#ifdef NILO_USE_TRACING
        dump_arp_packet (arp_reply);
        dump_arp_table (&arptable);
#endif
	  } else {
	    /* We have this entry, so ignore it */
	    NILO_TRACE_MSG ("We already have this entry.\r\n");
	    continue;
	  }
	  if (client_info->client_state == ARPING)
	    client_info->client_state = ARP_DONE;
	  //	  udpread->status = ;
	  NILO_TRACE_FUNCTION_RETURN;
	  return PXENV_EXIT_SUCCESS;
	}
	continue;
      }
	
      /* See if it's an IP packet */
      if ((packetlen < (ETHER_HDR_SIZE + sizeof(struct iphdr) +
			sizeof(struct udphdr))) ||
	  (ntohs((u_int16_t)packet[12]) == ETHERTYPE_IP)) {
	ip = (struct iphdr *)&packet[ETHER_HDR_SIZE];
#ifdef NILO_USE_TRACING_XXX
    dump_ip_packet (ip);
#endif  /* NILO_USE_TRACING  */
    if ((ip->verhdrlen != 0x45) ||
        dhcp_ipchksum(ip, sizeof(struct iphdr)) ||
        (ip->protocol == IP_UDP)) {
	
	/* if udpread->dest_ip is zero, read all packets */
	memcpy (&num, ip->dest, 4);
	if (udpread->dest_ip.num == UDP_READ_ANY_IP) {
            NILO_TRACE_MSG ("Received a packet for IP %s\r\n",
                            _inet_ntoa(num));
	} else {
            if (num == udpread->dest_ip.num) {
                NILO_TRACE_MSG ("Received a packet for the specified IP %s\r\n",
                                _inet_ntoa(num));
            } else {
                NILO_TRACE_MSG ("Unwanted packet for IP %s, %s desired.\r\n",
                                _inet_ntoa(num),
                                _inet_ntoa(udpread->dest_ip.num));
                continue;
            }
	}
        
	udp = (struct udphdr *)&packet[ETHER_HDR_SIZE + sizeof(struct iphdr)];
#ifdef NILO_USE_TRACING_XXX
        dump_udp_packet (udp);
#endif  /* NILO_USE_TRACING  */
        
	//	udpread->d_port = udp->dest;
	udpread->s_port = udp->src; /* set the source UDP port */
	//	udpread->buffer_size = udp->len;
	udpread->buffer = packet+ETHER_HDR_SIZE+sizeof(struct iphdr)+sizeof(struct udphdr);
	// memcpy ((char *)udpread->buffer, (char *)udp, udp->len);
	
	/* if udpread->d_port is zero, read all packets */
	if (udpread->d_port == UDP_READ_ANY_PORT) {
	  NILO_TRACE_MSG ("Received a packet on port 0x%02x\r\n",
			  ntohs(udp->dest));
	  udpread->d_port = udp->dest;
	  NILO_TRACE_FUNCTION_RETURN;
	  return PXENV_EXIT_SUCCESS;
	} else {
	  if (ntohs(udp->dest) ==  udpread->d_port) {
	    /* if udpread->d_port is non-zero, then filter all incoming
	       packets and receive only the ones for this UDP port. */
	    NILO_TRACE_MSG ("Received a packet on the specified port 0x%02x\r\n",
			    ntohs(udp->dest));
	    NILO_TRACE_FUNCTION_RETURN;
	    return PXENV_EXIT_SUCCESS;
	  } else {
	    NILO_TRACE_MSG ("Unwanted packet on port 0x%02x, wanted port 0x%02x.\r\n",
			    ntohs(udp->dest), udpread->d_port);
	    continue;
	  }
	}
      } else {
	NILO_TRACE_MSG ("Unidentified non-IP packet\r\n");
	continue;
      }
      }
      }
      //      if (!num)
      //	NILO_TRACE_MSG ("The packet queue is empty.\r\n");
    }
      udpread->buffer_size = 0;
      
      NILO_TRACE_FUNCTION_RETURN;
      return PXENV_EXIT_FAILURE;
}

/* pxenv_udp_write() - writes one packet to the requested UDP port
 *      using an already opened UDP connection.
 *
 * parameters:
 *     udpwrite.dest_ip - the IP number of the destination.
 *     udpwrite.gw_ip - the ip number of the gateway.
 *     udpwrite.src_port - the UDP port of the sender,
 *                         defaults to 2069 if this flag is zero.
 *     udpwrite.dest_port - the UDP port of the destination.
 *     udpwrite.buffer_size - size of the transmit buffer.
 *     udpwrite.buffer - the buffer address.
 *
 * returns:
 *     udpwrite.status - the exit status of this function. This is set
 *                       to one of the PXENV_STATUS_xxx constants.
 */
pxenv_ret_code_t
pxenv_udp_write (pxenv_udp_write_t *udpwrite)
{
//    char *buf = bufv;
    struct iphdr *ip;
    struct udphdr *udp;
    struct udp_pkt *full_pkt;
    pxenv_undi_tbd_t undi_tbd;
    pxenv_undi_transmit_t unditrans;
    struct arprequest arpreq;
    int arpentry, i;
    int retry = 6;
    u_int32_t temp;
 
    NILO_TRACE_FUNCTION;

    if ((full_pkt = (struct udp_pkt *)malloc(sizeof(struct udp_pkt))) == 0) {
        panic("Called malloc any memory for full_pkt");
    }
    
    memcpy (&full_pkt->u, udpwrite->buffer, udpwrite->buffer_size);
    udpwrite->buffer_size += sizeof(struct iphdr) + sizeof(struct udphdr);

    /* Add the version header length */
    full_pkt->ip.verhdrlen = 0x45;
    /* Add the type of the service */
    full_pkt->ip.service = 0;
    /* Add the size of the IP packet */
    full_pkt->ip.len = htons(udpwrite->buffer_size);
    /* FIXME: Ident should be random */
    full_pkt->ip.ident = 0;
    full_pkt->ip.frags = 0;
    full_pkt->ip.ttl = 60;              /* FIXME: was 64 */
    full_pkt->ip.protocol = IP_UDP;
    full_pkt->ip.chksum = 0;
    temp = udpwrite->gw_ip.num;
    memcpy(full_pkt->ip.src, &temp, 4);
    temp = udpwrite->dest_ip.num;
    memcpy(full_pkt->ip.dest, &temp, 4);
    full_pkt->ip.chksum = dhcp_ipchksum(&full_pkt->ip,
                                        sizeof(struct iphdr));
    
    full_pkt->udp.src = htons(udpwrite->src_port);
    full_pkt->udp.dest = htons(udpwrite->dest_port);
    full_pkt->udp.len = htons(udpwrite->buffer_size - sizeof(struct iphdr));
    /* FIXME: we should really calculate the checksum, but as
       it's optional, we don't bother for now. */
    full_pkt->udp.chksum = 0;
//    full_pkt->udp.chksum = dhcp_ipchksum(&full_pkt->udp, sizeof(struct udphdr));

#ifdef NILO_USE_TRACING_XXX
    dump_ip_packet (&full_pkt->ip);
    dump_udp_packet (&full_pkt->udp);
#endif  /* NILO_USE_TRACING  */

     unditrans.protocol = P_IP;
     /* if zero, use the dest_addr. If 1, the adapator supplies the
        broadcast address. */
     if (udpwrite->dest_ip.num == 0) {
       unditrans.xmit_flags = XMT_BROADCAST;
       memcpy (unditrans.dest_addr, bcastaddr, sizeof(bcastaddr));
     } else {
       unditrans.xmit_flags = XMT_DESTADDR;
       /* FIXME: get the hwaddr from the IP number */
       if (lookup_arp(udpwrite->dest_ip) == 0) {
	 fill_arp_entry (ARP_SERVER);
       }
       memcpy (unditrans.dest_addr, lookup_arp(udpwrite->dest_ip), sizeof(bcastaddr));
     }

     /* the hardware address of the destination media. If this isn't
        known, this field can be zero, and xmit_flags must be XMT_BROADCAST. */
     //     memset (unditrans.dest_addr, 0, sizeof(unditrans.dest_addr));
     //     memcpy (unditrans.dest_addr, bcastaddr, sizeof(bcastaddr));
     undi_tbd.length = udpwrite->buffer_size;
     undi_tbd.xmit_buffer = (addr32_t *)full_pkt;
     unditrans.tbd_addr = &undi_tbd;
     
     if (call_pxe_api (PXENV_UNDI_TRANSMIT, &unditrans) == PXENV_EXIT_FAILURE) {
         panic ("Couldn't transmit a packet on using the UNDI.\r\n");
         return (PXENV_EXIT_FAILURE);
     }

//    dhcp_eth_transmit(ether_src, full_pkt->ip.dest, ETHERTYPE_IP,
//                      len, full_pkt);

//    dhcp_eth_transmit(ether_src, ether_bcast, ETHERTYPE_IP, len, full_pkt);

    NILO_TRACE_FUNCTION_RETURN;

    return (PXENV_EXIT_SUCCESS);
}

