/* Nilo - A Network boot loader.

   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the TFTP data structures */
#include "tftp.h"

/* Define the main Nilo data structures */
#include "nilo.h"

/* Define the debug functions that can print the data structures. */
#include "dump.h"

#include <stdarg.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tftp.h"

static u_int32_t blockno = 0; 
// static u_int32_t nextblockno = 1;
// static int16_t opcode = RRQ;
static int32_t file_loaded = 0;

int
fill_tftp_request(tftp_t *tftp_pkt, u_int16_t opcode,
                  const char *filename, int32_t blockno)
{
    int len;

    NILO_TRACE_FUNCTION;

    switch (opcode) {
    case TFTP_RRQ:
        /* This is the max default size of a tftp packet. Intel's
           MTFTP uses a packet size of TFTP_MAX_PACKET, which being
           larger, speeds things up a bit. This size can be changed
           to an arbitrary size by a TFTP_OACK packet. */
      //        client_info->tftp_blockno++;
        client_info->tftp_blksize = TFTP_DEFAULTSIZE_PACKET;
        tftp_pkt->opcode = htons(TFTP_RRQ);
        if (filename[0] == '\0') {
            filename = TFTP_DEFAULT_FILENAME;
        } else {
            strcpy(tftp_pkt->u.rrq, filename);
        }
        len = strlen(filename);
        strcpy((tftp_pkt->u.rrq+(len+1)), "netascii");
        len += 1;                       /* add a count for the NULL */
        len += strlen("netascii");
        len += 1;                       /* add a count for the NULL */
        len += 2;                       /* add a count for the opcode */
        break;
    case TFTP_ACK:
        tftp_pkt->opcode = htons(TFTP_ACK);
        tftp_pkt->u.ack.block = htons(blockno);
        len = sizeof(short) * 2;
        break;
    default:
        NILO_TRACE_MSG ("Don't know how to fill with unknown packet type %d\n", opcode);
        break;
    }

#ifdef NILO_USE_TRACING
//    dump_tftp_packet (tftp_pkt);
#endif /* NILO_USE_TRACE */

    NILO_TRACE_FUNCTION_RETURN;
    
    return (len);
}

int
tftp_decode_packet(tftp_t *pkt, oskit_size_t pkt_size)
{
    u_int32_t i, j, len, endtag, num;
    /* FIXME: snaplen is bogus here */
    u_int32_t snaplen;
    u_int16_t status;
#ifdef	SIMULATE_BLOCK_LOSS
    static int32_t donelosing = 0;
#endif

    NILO_TRACE_FUNCTION;

#ifdef NILO_USE_TRACING
    dump_tftp_packet (pkt);
#endif /* NILO_USE_TRACING */

    snaplen -= client_info->tftp_blksize - sizeof(short) * 2;
    switch (status = ntohs(pkt->opcode)) {
    case TFTP_DATA:
      if (ntohs(pkt->u.data.block) == client_info->tftp_blockno) /* not a dup */
	handle_block(client_info->tftp_blockno,
		     pkt->u.data.download, snaplen);
      else if (ntohs(pkt->u.data.block) > client_info->tftp_blockno)
	/* skipped a block, fatal what do we do? */
	NILO_TRACE_MSG ("Skipped a TFTP data block, got 0x%02,"
			"expected 0x%02x\n", pkt->u.data.block,
			client_info->tftp_blockno);
      
      //        if (client_info->client_state == DHCP_DONE ||
      //            client_info->client_state == TFTP_REQUEST)
      client_info->client_state= TFTP_LOADING;
      //            client_info->client_state = TFTP_ACKING;
      if (pkt_size < client_info->tftp_blksize)
	client_info->client_state = TFTP_DONE;
      break;
    case TFTP_ERROR:
      NILO_TRACE_MSG ("ERROR %d: %s\n", ntohs(pkt->u.err.errcode),
		      pkt->u.err.errmsg);
      break;
    case TFTP_OACK:
      if (strcmp(pkt->u.oack.data, "blksize") == 0) {
	client_info->tftp_blksize = atol(pkt->u.oack.data
					 + sizeof("blksize"));
	NILO_TRACE_MSG ("OACK Setting TFTP block size to 0x%02x\n",
			client_info->tftp_blksize);
      }
      client_info->client_state = TFTP_ACKING;
      break;
    default:
      NILO_TRACE_MSG ("Unknown TFTP packet type 0x%02x\n", status);
      break;
    }
    //    client_info->tftp_blockno++;
    client_info->client_state = TFTP_ACKING;
    NILO_TRACE_FUNCTION_RETURN;
}





