/*
 * Copyright (c) 1997-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/* Modified from FreeBSD 2.1 sys/i386/boot/netboot */
/**************************************************************************
NETBOOT -  BOOTP/TFTP Bootstrap Program

Author: Martin Renters
  Date: Dec/93

Modified by: Rob Savoye <rob@welcomehome.org>
  Date: April/99
**************************************************************************/

/* This is the include generated by configuring */
#include "niloconfig.h"

#include <oskit/c/sys/types.h>
#include <oskit/net/ether.h>
#include <oskit/c/arpa/inet.h>
#include <oskit/c/string.h>
#include <oskit/c/assert.h>
#include <oskit/c/stdio.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

#include "udp.h"
#include "arp.h"
#include "tftp.h"
#include "dhcp.h"

/* The arptable is maintained in host order. */
//arptable_t arptable[MAX_ARP];

static char ether_bcast[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/*
 * IP checksum algorithm.
 */
u_int16_t
//dhcp_ipchksum(struct iphdr *ipv, int len)
dhcp_ipchksum(void *ipv, int len)
{
    u_int16_t *ip = (u_int16_t *)ipv;
    u_int32_t sum = 0;
    
    //    NILO_TRACE_FUNCTION;
    
    len >>= 1;
    while (len--) {
        sum += *(ip++);
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
    }

    // NILO_TRACE_FUNCTION_RETURN;

    return((~sum) & 0x0000FFFF);
}




